/*
 * Euro 2008 is Copyright (c) 2008 by Thomas Pundt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice and this paragraph and the
 * following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <PalmOS.h>
#include <DateGlue.h>

#include "euro2008.h"
#include "euro2008rcp.h"

#include "history.h"

typedef enum { tieBreakerPoints, 
	       tieBreakerGoalsDiff,
	       tieBreakerGoals,
	       tieBreakerDirectCompare,
	       tieBreakerUefaCoefficient
} tieBreakerType;

struct match match[8];

struct teamState teamState[4], backupTeamState[4];

MyAppPreferencesType appPrefs;
Int16 selectedMatch = -1;
UInt16 selectedTeam = 0;
Int16 selectedGroup = -1;
Int32 defaultDateFormat;
Int8 selectedYear = 0;

static TablePtr leagueTable;
Char* progDatePtr;

static void SetDateString(Int16 id, UInt32 date)
{
  DateTimeType dateP;
  ControlType *ctlP = Id2Ptr(id);
  char *label = (char*)CtlGetLabel(ctlP);
  CtlEraseControl(ctlP);

  TimSecondsToDateTime(date+appPrefs.gmtOff, &dateP);
  DateToAscii(dateP.month,dateP.day,dateP.year,
	      PrefGetPreference(prefDateFormat), label);

  // don't draw 0:00 time
  if (dateP.hour+dateP.minute>0) {
    StrCat(label," ");
    TimeToAscii(dateP.hour,dateP.minute,tfColon24h,
		label+StrLen(label));
  }
  CtlDrawControl(ctlP);
}

static void SetTimezoneString(Int16 id, int num, const Char* format)
{
  char *label;
  ControlType *ctlP = Id2Ptr(id);
  label = (char*)CtlGetLabel(ctlP);
  StrPrintF(label,format,num);
  CtlEraseControl(ctlP);
  CtlDrawControl(ctlP);
}

/*
 * I guess they'll handle it the same way in Euro 2008:
 * 
 * This comparison function takes into account items 6.a)-6.c) from
 * Article 28) of FIFA Worldcup 2002 Reglement or items 6.d)-6.g)
 * (points, goals and goals difference equality) if executed from a
 * call to ComputeTable (with a smaller table).
 */
static char cmpTeamsByTiebreaker(struct teamState *a, struct teamState *b, UInt8 t)
{
  switch (t) {
  case tieBreakerPoints:
    if (a->Points > b->Points) return 1;
    if (a->Points < b->Points) return -1;
    break;
  case tieBreakerGoalsDiff:
    if (a->Goals + b->GoalsAgainst > b->Goals + a->GoalsAgainst) return 1;
    if (a->Goals + b->GoalsAgainst < b->Goals + a->GoalsAgainst) return -1;
    break;
  case tieBreakerGoals:
    if (a->Goals > b->Goals) return 1;
    if (a->Goals < b->Goals) return -1;
    break;
  case tieBreakerUefaCoefficient:
    if (a->teamNo < b->teamNo) return 1;
    if (a->teamNo > b->teamNo) return -1;
    break;
  }
  return 0;
}

static Int16 cmpByPosition(void *a, void *b, Int32 other)
{
  if (((struct teamState*)a)->Position > ((struct teamState*)b)->Position) return 1;
  if (((struct teamState*)a)->Position < ((struct teamState*)b)->Position) return -1;
  return 0;
}

static void AddTeamState(struct teamState *table, Int16 n, Int16 n1, Int16 n2)
{
  Int16 pkte = 3;
  if (n>0) n--;

  table[n].Goals  += n1;
  table[n].GoalsAgainst += n2;
  table[n].Matches++;
  if (n1>n2) {
    table[n].Points+=pkte;
    table[n].Won++;
  } else if (n1==n2) {
    table[n].Points++;
    table[n].Draw++;
  } else {
    table[n].Lost++;
  }
}

/*
 * compute the positions where in the table the ranking changes
 */
static UInt32 computePartition(struct teamState *table)
{
  Int8 prevPosition = -1;
  UInt32 partition = 0, i;

  SysInsertionSort(table, 4, sizeof(struct teamState),
		   cmpByPosition, 0);
  for (i=0; i<4; i++) {
    if (table[i].Position != prevPosition) {
      partition = partition | (1UL<<i);
      prevPosition = table[i].Position;
    }
  }
  return partition;
}

/*
 * compute table;
 * table:     standings table
 * match:     matches array
 * teamMask:  bit field of teams to compute table for
 * returns:   "final" flag.
 */
Int16 ComputeTable(struct teamState *table, struct match *match, UInt32 teamMask)
{
  Int16 i,j,t,final=1;
  UInt32 partition;
  Int16 partitionStartPoint, partitionEndPoint;
  tieBreakerType tiebreaker[5] = { 
    tieBreakerPoints, 
    tieBreakerDirectCompare,
    tieBreakerGoalsDiff,
    tieBreakerGoals,
    tieBreakerUefaCoefficient
  };

  MemSet(table,4*sizeof(struct teamState),0);

  // first step: reset previous values to zero.
  for (i=0; i<4; i++) {
    table[i].teamNo = i+1;
  }

  // second step: loop over all records necessary and sum up
  // goals, points and handicaps.


  // take every match of every team within "who"
  for (i=0; i<6; i++) {
    if (((1<<(match[i].team1-1)) & teamMask) 
	&& ((1<<(match[i].team2-1)) & teamMask) ) {
      if ( (match[i].erg1 != -1) && (match[i].erg2 != -1) ) {
	AddTeamState(table,match[i].team1,match[i].erg1,match[i].erg2);
	AddTeamState(table,match[i].team2,match[i].erg2,match[i].erg1);
      } else {
	// there is at least one match missing for a final table.
	final = 0;
      }
    }
  }


  partition = 1; // initial partition of table

  // we do table/partition sorting for each tiebreaker
  for (t=0; t<sizeof(tiebreaker)/sizeof(tieBreakerType); t++) {

    // compute start and end position of a partition;
    // initial partition is the whole table.
    partitionStartPoint = 0;
    for ( partitionEndPoint = partitionStartPoint+1; 
	  partitionEndPoint <= 4; 
	  partitionEndPoint++ ) {
      if ( (partition & (1UL<<partitionEndPoint)) || 
	   partitionEndPoint == 4) {

	// now sort the partition by selected tiebreaker.

	// handle "direct encounter" differently, because a separate
	// "sub table" has to be built with all teams involved.
	if (tiebreaker[t] == tieBreakerDirectCompare 
	    && partitionEndPoint-partitionStartPoint > 1) { // direct comparison 
	      
	  UInt32 mask = 0;

	  // if we have recursively come to this point, there is
	  // not much to check. Procede with next tiebreaker.
	  if (teamMask != 0xffffffff) 
	    continue;

	  // build mask of involved teams
	  for ( i=partitionStartPoint; i<partitionEndPoint; i++) {
	    mask = mask | (1UL<< (table[i].teamNo-1));
	  }

	  // save TeamResult table for later reference
	  MemMove(backupTeamState, table, 4*sizeof(struct teamState));

	  // compute sub table, evaluate results
	  ComputeTable(table, match, mask);
	  for ( i=partitionStartPoint; i<partitionEndPoint; i++) {
	    for (j=0; j<4; j++) {
	      if (backupTeamState[i].teamNo == table[j].teamNo)
		backupTeamState[i].Position += table[j].Position;
	    }
	  }

	  // restore original "corrected" TeamResult.
	  MemMove(table, backupTeamState, 4*sizeof(struct teamState));

	} else {
	  char sort;
	  // if we have recursively come to this point, don't
	  // check for UEFA coefficient - in tied matches this would
          // force a decision before having checked "global" tiebreakers.
	  if (teamMask != 0xffffffff && tiebreaker[t] == tieBreakerUefaCoefficient) 
	    continue;
	  for ( i=partitionStartPoint; i<partitionEndPoint-1; i++) {
	    if ( (teamMask & (1UL<<(table[i].teamNo-1))) ) {
	      for (j=i+1; j<partitionEndPoint; j++) {
		if ( (teamMask & (1UL<<(table[j].teamNo-1))) ) { 
		  sort = cmpTeamsByTiebreaker(&table[i],
					      &table[j],
					      tiebreaker[t]);
		  if (sort<0) table[i].Position++;
		  if (sort>0) table[j].Position++;
		}
	      }
	    }
	  }
	}

	partitionStartPoint = partitionEndPoint;
      }
    }

    partition = computePartition(table);
      
  }

  return final;
}

/**
 * Form Compute Table Functions
 */

static void DrawTableItem(void* tableP, Int16 row, Int16 column,
			  RectangleType *bounds)
{
  Int16 n;
  UInt16 x;
  Char buf[25];
  struct team *team;

  n = TblGetItemInt(tableP, row, column);
  x = bounds->topLeft.x;

  StrIToA(buf,n);

  switch (column) {
  case 0:
    StrCat(buf,".");
    break;
  case 1:
    team = GetTeamString((appPrefs.group-1)*4+n);
    MemMove(buf,team->name,StrLen(team->name)+1);
    break;
  case 3:
    StrCat(buf," :");
    x += bounds->extent.x - FntCharsWidth(buf,StrLen(buf)) - 1;
    break;
  }

  FntSetFont(0);
  WinEraseRectangle(bounds, 0);
  WinDrawChars(buf, StrLen(buf), x, bounds->topLeft.y);
}

/*
 * fills in the values for the table and redraws it.
 */
static void InitTable(void)
{
  Int16 n;

  for (n=0; n<4; n++) {
    TblSetItemInt(leagueTable, n, 0, teamState[n].Position+1);
    TblSetItemInt(leagueTable, n, 1, teamState[n].teamNo);
    TblSetItemInt(leagueTable, n, 2, teamState[n].Matches);
    TblSetItemInt(leagueTable, n, 3, teamState[n].Goals);
    TblSetItemInt(leagueTable, n, 4, teamState[n].GoalsAgainst);
    TblSetItemInt(leagueTable, n, 5, teamState[n].Points);
  }
}

/*
 * get the contents of the editable fields in the form and save them
 */
static void SaveMatchdayResults(void)
{
  Int16 i;
  char *e1, *e2;
  Int16 buf[2], rows;

  for (i=0; i<8; i++) {
    e1 = FldGetTextPtr(match[i].e1);
    e2 = FldGetTextPtr(match[i].e2);

    buf[0] = buf[1] = 255;

    if (e1 && StrLen(e1)
	&& e2 && StrLen(e2)) {
      buf[0] = StrAToI(e1);
      buf[1] = StrAToI(e2);
    }
    match[i].erg1 = buf[0];
    match[i].erg2 = buf[1];
  }

  rows = appPrefs.group>0 ? 6 : 7;
  WriteFixture(match, rows, appPrefs.group);
}

void ReadSetFixture(struct match* match, UInt16 group, Boolean setallfields)
{
  struct matchRecord matchRec;
  char *p;
  UInt16 i = 0, recNo;

  for (p=ReadFixture(0, &matchRec),recNo=0; 
       p; 
       p=ReadFixture(p, &matchRec),recNo++) {
    if (matchRec.group != group)
      continue;

    match[i].date  = matchRec.date;
    match[i].venue = matchRec.venue;
    match[i].erg1  = matchRec.e1;
    match[i].erg2  = matchRec.e2;
    match[i].team1 = matchRec.m1;
    match[i].team2 = matchRec.m2;
    match[i].recNo = recNo;
    if (setallfields) {
      if (group>0) {
	SetHandleString(match[i].m1, (group-1)*4 + matchRec.m1-1);
	SetHandleString(match[i].m2, (group-1)*4 + matchRec.m2-1);
	match[i].team1Final = matchRec.m1+(group-1)*4;
	match[i].team2Final = matchRec.m2+(group-1)*4;
      } else {
	match[i].team1Final = ComputeSetHandleString(match[i].m1, matchRec.m1);
	match[i].team2Final = ComputeSetHandleString(match[i].m2, matchRec.m2);
	SetHandleInt(match[i].num,recNo+1);
      }
      SetHandleInt(match[i].e1,matchRec.e1);
      SetHandleInt(match[i].e2,matchRec.e2);
    }
    i++;
  }
}

Int16 InitFormMain(void)
{
  Int16 i, j, rows;
  Int16 final = 0;
  FormType* frm = FrmGetActiveForm();
  // const char* p;

  selectedMatch = -1;
  selectedTeam = 0;
  selectedGroup = appPrefs.group;

  rows = appPrefs.group > 0 ? 6 : 7;

  /*
   * initialize table variables to default values;
   */
  leagueTable = Id2Ptr(IdFrmMain+100);

  for (i=0; i<8; i++) {
    match[i].m1 = Id2Ptr(IdFrmMain+20+i);
    match[i].m2 = Id2Ptr(IdFrmMain+30+i);
    match[i].e1 = Id2Ptr(IdFrmMain+40+i);
    match[i].e2 = Id2Ptr(IdFrmMain+50+i);
    match[i].num = Id2Ptr(IdFrmMain+60+i);
    FldSetUsable(match[i].m1, i<rows?true:false);
    FldSetUsable(match[i].m2, i<rows?true:false);
    FldSetUsable(match[i].e1, i<rows?true:false);
    FldSetUsable(match[i].e2, i<rows?true:false);
    FldSetUsable(match[i].num, i<rows?true:false);
    CtlSetUsable(Id2Ptr(IdFrmMain+10+i), i<rows?true:false);
    if (appPrefs.group == 0) {
      Int16 fixup = 0;
      if (i>3) fixup = 5;
      if (i>5) fixup = 10;
      for (j=10; j<70; j+=10) {
	FrmMoveObject(IdFrmMain+j+i,0,fixup);
      }
    }
  }

  ReadSetFixture(match,appPrefs.group,true);
  if (appPrefs.group>0) {
    for (i=0; i<4; i++) {
      TblSetRowSelectable(leagueTable, i, false);
      TblSetRowUsable(leagueTable, i, appPrefs.group>0?true:false);
      for (j=0; j<6; j++) {
	TblSetItemStyle(leagueTable, i, j, customTableItem);
	TblSetColumnUsable(leagueTable, j, true);
	TblSetCustomDrawProcedure(leagueTable, j, DrawTableItem);
      }
    }
    final = ComputeTable(teamState,match,0xffffffff); 
    InitTable();
  } else {
    // FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+2));
    FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+7));
    FrmHideObject(frm, FrmGetObjectIndex(frm,IdFrmMain+8));
    FrmSetObjectPosition(frm, FrmGetObjectIndex(frm,IdFrmMain+100),0,159);
  }
  FrmCopyLabel(frm, IdFrmMain,
	       LstGetSelectionText(Id2Ptr(IdFrmMain+1),appPrefs.group));
  // p = LstGetSelectionText(Id2Ptr(IdFrmMain+1),appPrefs.group);
  // MemMove((char*)CtlGetLabel(Id2Ptr(IdFrmMain)),p,StrLen(p)+1);
  FrmDrawForm(frm);

  DrawButton(IdFrmMain+91,IdBitmap+11);
  DrawButton(IdFrmMain+92,IdBitmap+12);
  DrawButton(IdFrmMain+93,IdBitmap+14);

  return final;
}

Boolean MatchdayFormHandleEvent(EventPtr e)
{
  Boolean handled = false;
  Int16 item;
  ListPtr pLst;
  static Int16 final;
  static Boolean historyMode = false;

  switch (e->eType) {

  case frmOpenEvent:
    final = InitFormMain();
    handled = true;
    break;

  case menuEvent:

    switch(e->data.menu.itemID) {

    case IdMenuItem:
      FrmPopupForm(IdFrmAbout);
      break;

    case IdMenuItem+1:
      FrmHelp(IdFrmMain);
      break;

    case IdMenuItem+2:
      FrmPopupForm(IdFrmTimezone);
      break;

    case IdMenuItem+4:
      SaveMatchdayResults();
      FrmGotoForm(IdFrmMatches);
      break;

    case IdMenuItem+5:
      BeamDatabase();
      break;

    case IdMenuItem+6:
      SaveMatchdayResults();
      FrmGotoForm(IdFrmFixture);
      break;

    case IdMenuItem+7:
      historyMode = true;
      break;

    }
    handled = true;
    break;

  case ctlSelectEvent:
    switch (e->data.ctlEnter.controlID) {

    popup:
    case IdFrmMain:
      pLst = Id2Ptr(IdFrmMain+ (appPrefs.options & OPT_AlternativeGroupLabels?9:1));
      LstSetSelection(pLst, appPrefs.group);
      LstMakeItemVisible(pLst, appPrefs.group);
      LstPopupList(pLst);
      if ( (item = LstGetSelection(pLst)) != appPrefs.group && item!=-1) {
	SaveMatchdayResults();
	appPrefs.group = item;
	FrmGotoForm(IdFrmMain);
      }
      handled = true;
      break;

    case IdFrmMain+2: /* Down-Arrow => Compute Table */
      SaveMatchdayResults();
      if (appPrefs.group > 0) {
	final = ComputeTable(teamState,match,0xffffffff);
	InitTable();
	TblDrawTable(leagueTable);
      } else {
	ReadSetFixture(match,appPrefs.group,true);
	FrmDrawForm(FrmGetActiveForm());
      }
      handled = true;
      break;

    case IdFrmMain+92: // "List form" button
      SaveMatchdayResults();
      FrmGotoForm(IdFrmMatches);
      break;

    case IdFrmMain+93: // "Overview form" button
      SaveMatchdayResults();
      FrmGotoForm(IdFrmFixture);
      break;

    default:
      /* (i) button => show popup with venue and date of match */
      if (e->data.ctlEnter.controlID >= IdFrmMain+10 &&
	  e->data.ctlEnter.controlID <= IdFrmMain+17) {
	int index = e->data.ctlEnter.controlID - IdFrmMain-10;
	selectedMatch = index;
	FrmPopupForm(IdFrmMainPopup);
	handled = true;
      }
    }
    break;

  case fldEnterEvent:
    item = e->data.fldEnter.fieldID-IdFrmMain-20;
    if (item>=0 && item<8) {
      handled = true;
      selectedTeam = match[item].team1Final;
    } else if (item>=10 && item<18) {
      handled = true;
      item -= 10;
      selectedTeam = match[item].team2Final;
    }
    if (handled && selectedTeam>0 && selectedTeam<=32)
      FrmPopupForm(IdFrmTeams);
    else if (e->data.fldEnter.fieldID >= IdFrmMain+40 &&
	     e->data.fldEnter.fieldID <= IdFrmMain+57) {
      // Compute table whenever you enter another field.
      FormPtr frm = FrmGetActiveForm();
      SaveMatchdayResults();
      final = ComputeTable(teamState,match,0xffffffff);
      InitTable();
      TblDrawTable(leagueTable);
      FrmSetFocus(frm, FrmGetObjectIndex(frm,e->data.fldEnter.fieldID));
    }
    break;

  case keyDownEvent:
    if (e->data.keyDown.chr == vchrPageUp) {
      SaveMatchdayResults();
      appPrefs.group--;
      if (appPrefs.group < 0) appPrefs.group = 4;
      FrmGotoForm(IdFrmMain);
      handled = true;
    }
              
    if (e->data.keyDown.chr == vchrPageDown) {
      SaveMatchdayResults();
      appPrefs.group++;
      if (appPrefs.group > 4) appPrefs.group = 0;
      FrmGotoForm(IdFrmMain);
      handled = true;
      break;
    }

    if (e->data.keyDown.chr == vchrRockerLeft ||
	e->data.keyDown.keyCode == fiveWayLeft) {
      FrmGotoForm(IdFrmFixture);
      handled = true;
    }
    
    if (e->data.keyDown.chr == vchrRockerRight || 
	e->data.keyDown.keyCode == fiveWayRight) {
      FrmGotoForm(IdFrmMatches);
      handled = true;
    }

    if (e->data.keyDown.chr == vchrRockerCenter || 
	(e->data.keyDown.keyCode == fiveWaySelect && 
	 (e->data.keyDown.modifiers & autoRepeatKeyMask) == 0) ) {
      goto popup;
    }

    break;

  default:
    break;
  }

  return handled;
}

Boolean MainPopupFormHandleEvent(EventPtr e)
{
  Boolean handled = false;

  switch (e->eType) {
  case frmOpenEvent:
    // display the team names
    if (selectedGroup>0) { // first round
      SetHandleString(Id2Ptr(IdFrmMainPopup), 
                      (appPrefs.group-1)*4 + match[selectedMatch].team1 - 1);
      SetHandleString(Id2Ptr(IdFrmMainPopup + 1), 
                      (appPrefs.group-1)*4 + match[selectedMatch].team2 -1);
    } else { // group == 0 --> second round
      ComputeSetHandleString(Id2Ptr(IdFrmMainPopup), 
                             match[selectedMatch].team1);
      ComputeSetHandleString(Id2Ptr(IdFrmMainPopup + 1), 
                             match[selectedMatch].team2);
    }

    FrmDrawForm(FrmGetActiveForm());
    // display stadium
    SetLocationString(IdFrmMainPopup+2, match[selectedMatch].venue-1);
    // display time of match
    SetDateString(IdFrmMainPopup+3, match[selectedMatch].date);

    handled = true;
    break;

  case fldEnterEvent:
    if ( e->data.fldEnter.fieldID == IdFrmMainPopup+2 ) {
      if (selectedMatch != -1)
	FrmPopupForm(IdFrmStadium);
    }
    handled = true;
    break;

  case ctlSelectEvent:
    if (e->data.ctlEnter.controlID == IdFrmMainPopup+3 ) {
      if ( selectedMatch != -1 &&
	   0 == DisplayConfirmationMessage(IdErrorMsg+4)) {
	if (OpenDatebook()) {
	  MakeDatebookEntry(match[selectedMatch].date + appPrefs.gmtOff,
			    (match[selectedMatch].team1Final<<8)+
			    match[selectedMatch].team2Final);
	  DmCloseDatabase(datebook);
	}
      }
      handled = true;
      break;
    }
    break;

  case penDownEvent:
    if (e->screenY > 27) { // User taps outside the popup --> dismiss it
      // 27 = hard coded form height. Should be obtained dynamically.
      FrmReturnToForm(0);
      handled = true;
    }
    break;

  default:
    break;
  }

  return handled;
}

/**
 * Credits Form (via (i) in About Form)
 */
static void InitFrmHelp(void)
{
  FormPtr frm = FrmGetActiveForm();
  FieldPtr fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, IdFrmHelp));
  static char* msg;
  msg = GetString(IdFrmAbout);
  FldSetTextPtr(fld, msg);
  FldRecalculateField(fld, true);
  FldRecalculateScrollbar(fld, Id2Ptr(IdFrmHelp+1));
}

static Boolean FrmHelpHandleEvent (EventPtr e)
{
  Boolean handled = false;
  
  switch (e->eType) {
  case sclRepeatEvent:
    FldDoScrolling(Id2Ptr(IdFrmHelp), Id2Ptr(IdFrmHelp+1));
    break;
  
  default:
    ;
  }
  return handled;
}

/**
 * About Form
 */
Boolean AboutFormHandleEvent(EventPtr e)
{
  Boolean handled;
  FormPtr frm;
  FieldPtr fld;

  handled = false;
  switch (e->eType) {

  case frmOpenEvent:
    if (!progDatePtr) {
      Int16 len;
      len = DateGlueTemplateToAscii("^0r. ^2l ^4r",
				    tIntProgDateMonth,
				    tIntProgDateDay,
				    tIntProgDateYear,
				    0,0);
      progDatePtr = (Char*)MemPtrNew(len+1);
      DateGlueTemplateToAscii("^0r. ^2l ^4r",
			      tIntProgDateMonth,
			      tIntProgDateDay,
			      tIntProgDateYear,
			      progDatePtr,len+1);
    }
    frm = FrmGetActiveForm();
    fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, IdFrmAbout+1));
    FldSetTextPtr(fld, progDatePtr);
    FrmDrawForm(FrmGetActiveForm());
    handled = true;
    break;

  case ctlSelectEvent:
    if (e->data.ctlSelect.controlID == IdFrmAbout) {
      FrmReturnToForm(0);
      handled = true;
    }
    break;

  case penDownEvent:
    // this is the position of the (i)
    if (e->screenX >= 144 && e->screenX <= 155 &&
	e->screenY >= 0 && e->screenY <= 11) {
      // this way we don't have to use the default bold font.
      MyDoDialog(IdFrmHelp, FrmHelpHandleEvent, InitFrmHelp);
      handled =true;
    }

  default:
    break;
  }

  return handled;
}

/**
 * Timezone and Options Form
 */
Boolean TimezoneFormHandleEvent(EventPtr e)
{
  static UInt16 element;
  static Int32 newGmtOff, minutes;
  static Boolean timezoneChanged;
  Boolean handled;
  FormPtr frm;

  handled = false;
  switch (e->eType) {

  case frmOpenEvent:
    element = 0;
    timezoneChanged = false;
    // just to be sure, that the date really is correctly "aligned"
    newGmtOff = (appPrefs.gmtOff/1800L)*1800L;
    minutes = (newGmtOff%3600L)/60L;
    frm = FrmGetActiveForm();
    FrmSetControlGroupSelection(frm, 1, IdFrmTimezone+2);
    FrmSetControlValue(frm, FrmGetObjectIndex(frm, IdFrmTimezone+6),
		       (appPrefs.options & OPT_UseFrmMatches));
    FrmSetControlValue(frm, FrmGetObjectIndex(frm, IdFrmTimezone+7),
		       (appPrefs.options & OPT_AlternativeGroupLabels));

    FrmDrawForm(frm);
    SetTimezoneString(IdFrmTimezone+2,newGmtOff/3600L,"%+2d");
    SetTimezoneString(IdFrmTimezone+3,minutes,"%02d");

    if (FeatureSet35) {
      WinPushDrawState();
      WinSetForeColor(UIColorGetTableEntryIndex(UIDialogFrame));
    }
    WinDrawLine(0,46,159,46);
    WinDrawLine(0,47,159,47);
    if (FeatureSet35) {
      WinPopDrawState();
    }

    handled = true;
    break;

  case ctlSelectEvent:
    switch (e->data.ctlSelect.controlID) {
    case IdFrmTimezone:
      appPrefs.gmtOff = newGmtOff;
    case IdFrmTimezone+1:
      FrmReturnToForm(0);
      if (timezoneChanged)
        FrmGotoForm(FrmGetFormId(FrmGetActiveForm()));
      handled = true;
      break;

    case IdFrmTimezone+2:
      element = 0;
      handled = true;
      break;
    case IdFrmTimezone+3:
      element = 1;
      handled = true;
      break;

    case IdFrmTimezone+4:
      handled = true;
      timezoneChanged = true;
      /* according to a time zone map there really _is_ a time zone
       * that's off by +1400...
       */
      if (element == 0) {
	if (newGmtOff<=13L*3600L) {
	  newGmtOff += 3600L;
	  SetTimezoneString(IdFrmTimezone+2,newGmtOff/3600L,"%+2d");
	}
	break;
      }
    case IdFrmTimezone+5:
      handled = true;
      timezoneChanged = true;
      if (element == 1) { // doesn't matter if down or up...
	newGmtOff -= 60L*minutes;
	minutes = (30-minutes);
	newGmtOff += 60L*minutes;
	SetTimezoneString(IdFrmTimezone+3,minutes,"%02d");
	break;
      }
      if (element == 0 && newGmtOff>=-11L*3600L) {
	newGmtOff -= 3600L;
	SetTimezoneString(IdFrmTimezone+2,newGmtOff/3600L,"%+2d");
	break;
      }
      break;

    // the new options checkboxes
    case IdFrmTimezone+6:
      appPrefs.options ^= OPT_UseFrmMatches;
      handled = true;
      break;

    case IdFrmTimezone+7:
      appPrefs.options ^= OPT_AlternativeGroupLabels;
      handled = true;
      break;

    }
    break;

  default:
    break;
  }

  return handled;
}

